//
// simple_stats.hpp
//
// Bill Seymour, 2023-09-16
//
// Copyright Bill Seymour 2023.
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)
//
// This small template library provides a mechanism for doing some fairly
// abcedarian statistics:  just finding the minimum, maximum, median, mean,
// variance and standard deviation of numerical elements in a container
// that works like a std::vector or std::deque.
//
// It's intended to be simple and direct to use if the container elements
// are fundamental numeric types; but it's also usable, maybe with a bit
// of extra work on the part of the user, for user-defined types like,
// say, a rational number.
//
// A precondition for use is that the container be non-empty and already
// sorted ascending.
//

#ifndef SIMPLE_STATS_HPP_INCLUDED
#define SIMPLE_STATS_HPP_INCLUDED

#include <type_traits>
#include <tuple>
#include <utility> // move
#include <cmath>   // round, sqrt

namespace simpstat {

//
// Given the sorted precondition, min. and max. are trivial.
//
template<typename ContainerT> auto minimum(const ContainerT& data)
{
    return data.front();
}
template<typename ContainerT> auto maximum(const ContainerT& data)
{
    return data.back();
}

//
// For built-in numeric types, we'll have some defaults that
// will just work without the user having to do anything special.
// For user-defined types, users might have to define some or all
// of the following.
//
namespace dflt {

//
// Most of the work will actually be done using a "fractional type"
// that can hold values between adjacent (possibly integer) value types.
// We'll make it as big as we can hoping that sums of squares in the
// variance and standard deviation functions won't overflow.
//
typedef long double DefaultFrac;

//
// For narrowing conversions from the fractional type to the value type,
// we'll normally take whatever the standard conversion are.
//
template<typename ValT,
         typename FracT,
         bool FI /* floating-to-integer */,
         bool BI /* built-in FracT */>
struct defcast
{
    static ValT cast(FracT val)
    {
        return static_cast<ValT>(val);
    }
};
//
// But if we're converting from a built-in floating point type to a
// built-in integer type, we'll probably want to round to nearest.
//
template<typename ValT, typename FracT>
struct defcast<ValT,FracT,true,true>
{
    static ValT cast(FracT val)
    {
        return static_cast<ValT>(std::round(val));
    }
};
//
// And if the value type and the fractional type are the same type,
// we're almost certainly dealing with user-defined types, so we'll
// do a copy or a move as appropriate.
//
template<typename ValT>
struct defcast<ValT,ValT,false,false>
{
    static ValT cast(const ValT& val)
    {
        return val;
    }
    static ValT cast(ValT&& val)
    {
        return std::move(val);
    }
};

//
// A type trait for identifying narrowing conversions
// from a built-in fractional type to a built-in value type:
//
template<typename ValT, typename FracT>
struct FltToInt
{
    static constexpr bool value =
      std::is_integral<ValT>::value && std::is_floating_point<FracT>::value;
};

//
// Picking the correct specialization:
//
template<typename ValT, typename FracT>
using DefaultCaster = defcast<ValT,
                              FracT,
                              FltToInt<ValT,FracT>::value,
                              std::is_arithmetic<FracT>::value>;

//
// We'll need to extract a square root for computing the standard deviation,
// so we'll provide the following as a convenience in case it's good enough
// for user-defined types.
//
template<typename FracT, int ErrV = 1000, double FixedErrV = 0.0>
class NewtonSqrt
{
    //
    // Note that we require an explicit conversion from int to FracT.
    // We hope that it's a constant expression; but since we're dealing
    // with a user-defined type, we can't count on it.
    //
    static const FracT zero = static_cast<FracT>(0);
    static const FracT one  = static_cast<FracT>(1);
    static const FracT two  = static_cast<FracT>(2);
    static const FracT err  = static_cast<FracT>(ErrV);

    static FracT absval(FracT&& diff)
    {
        return std::move(diff < zero ? -diff : diff);
    }

public:
    static FracT sqrt(const FracT& val)
    {
        FracT error = INT_MIN == ErrV ? FracT(FixedErrV) : val / err;
        FracT root = one;
        while (error < absval(val - root * root))
        {
            // Newton's method:
            root = (val / root + root) / two;
        }
        return std::move(root);
    }
};

//
// For the square root of built-in floating point types, std::sqrt()
// is just what we need; otherwise, we'll default to NewtonSqrt above
// with the second and third template arguments defaulting.
//
template<typename FracT, bool BI>
struct defroot
{
    static FracT sqrt(FracT val)
    {
        return std::sqrt(val);
    }
};
template<typename FracT>
struct defroot<FracT,false>
{
    static FracT sqrt(const FracT& val)
    {
        return std::move(NewtonSqrt<FracT>::sqrt(val));
    }
};
template<typename FracT>
using DefaultRoot = defroot<FracT,std::is_floating_point<FracT>::value>;

} // namespace dflt

//
// Most of the actual work:
//
namespace frac {

template<typename ContainerT, typename FracT = dflt::DefaultFrac>
FracT median(const ContainerT& data)
{
    std::size_t siz = data.size();
    std::size_t mid = siz / 2;
    return (siz & 1) ? FracT(data[mid])
                     : FracT(data[mid-1] + data[mid]) / FracT(2);
}

template<typename ContainerT, typename FracT = dflt::DefaultFrac>
FracT mean(const ContainerT& data)
{
    typename ContainerT::value_type tot = typename ContainerT::value_type(0);
    for (const typename ContainerT::value_type& val : data)
    {
        tot += val;
    }
    return FracT(tot) / FracT(data.size());
}

//
// NB:  pass the already computed mean as a 2nd argument, same for stddev().
//
template<typename ContainerT, typename FracT = dflt::DefaultFrac>
FracT variance(const ContainerT& data, FracT avg)
{
    FracT tot = FracT(0);
    for (const typename ContainerT::value_type& val : data)
    {
        FracT diff = FracT(val) - avg;
        tot += diff * diff;
    }
    return tot / FracT(data.size());
}

template<typename ContainerT,
         typename FracT = dflt::DefaultFrac,
         typename RootT = dflt::DefaultRoot<FracT>>
FracT stddev(const ContainerT& data, FracT avg)
{
    return RootT::template sqrt(frac::variance<ContainerT,FracT>(data, avg));
}

//
// We might want to compute the mean, variance and standard deviation
// in a single function if we don't want to compute the mean separately
// and pass it to the variance and stddev functions.  Indeed, if we want
// both the variance and standard deviation, we won't want to go through
// the variance loop twice.
//
template<typename ContainerT,
         typename FracT = dflt::DefaultFrac,
         typename RootT = dflt::DefaultRoot<FracT>>
auto meanvardev(const ContainerT& data)
{
    FracT avg = frac::mean<typename ContainerT::value_type,FracT>(data);
    FracT var =
        frac::variance<typename ContainerT::value_type,FracT>(data, avg);
    FracT dev = RootT::template sqrt(var);
    return std::make_tuple(avg, var, dev);
}

} // namespace frac

//
// Users will probably want to deal with their value types
// and not worry about what the fractional type is.
//

template<typename ContainerT,
         typename FracT = dflt::DefaultFrac,
         typename CastT =
             dflt::DefaultCaster<typename ContainerT::value_type,FracT>>
auto median(const ContainerT& data)
{
    return CastT::cast(frac::median<ContainerT,FracT>(data));
}

template<typename ContainerT,
         typename FracT = dflt::DefaultFrac,
         typename CastT =
             dflt::DefaultCaster<typename ContainerT::value_type,FracT>>
auto mean(const ContainerT& data)
{
    return CastT::cast(frac::mean<ContainerT,FracT>(data));
}

template<typename ContainerT,
         typename FracT = dflt::DefaultFrac,
         typename CastT =
             dflt::DefaultCaster<typename ContainerT::value_type,FracT>>
auto variance(const ContainerT& data)
{
    return CastT::cast(frac::variance<ContainerT,FracT>
        (data, frac::mean<ContainerT,FracT>(data)));
}

template<typename ContainerT,
         typename FracT = dflt::DefaultFrac,
         typename RootT = dflt::DefaultRoot<FracT>,
         typename CastT =
             dflt::DefaultCaster<typename ContainerT::value_type,FracT>>
auto stddev(const ContainerT& data)
{
    return CastT::cast(frac::stddev<ContainerT,FracT,RootT>
        (data, frac::mean<ContainerT,FracT>(data)));
}

//
// The mean, variance and standard deviation in one fell swoop:
//
template<typename ContainerT,
         typename FracT = dflt::DefaultFrac,
         typename RootT = dflt::DefaultRoot<FracT>,
         typename CastT =
             dflt::DefaultCaster<typename ContainerT::value_type,FracT>>
auto meanvardev(const ContainerT& data)
{
    FracT avg = frac::mean<ContainerT,FracT>(data);
    FracT var = frac::variance<ContainerT,FracT>(data, avg);
    FracT dev = RootT::sqrt(var);
    return std::make_tuple
      (CastT::cast(std::move(avg)),
       CastT::cast(std::move(var)),
       CastT::cast(std::move(dev)));
}

} // namespace simpstat

#endif // SIMPLE_STATS_HPP_INCLUDED
